import json
import unittest
from datetime import datetime, timedelta
from string import Template

from json_timeseries import TsRecord, TimeSeries, JtsDocument


class TestTimeSeries(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        self.NOW = datetime.now()  # new Date(Math.round(new Date().getTime() / 1000) * 1000)
        self.ONE_MINUTE_AGO = self.NOW - timedelta(minutes=1)  # new Date(NOW.getTime() - (60 * 1000))
        self.TWO_MINUTE_AGO = self.NOW - timedelta(minutes=2)

        self.NUMBER_RECORDS = [
            TsRecord(**{"timestamp": self.ONE_MINUTE_AGO, "value": 1, "quality": 192, "annotation": 'comment'}),
            TsRecord(**{"timestamp": self.NOW, "value": 2}),
            TsRecord(**{"timestamp": self.TWO_MINUTE_AGO, "value": 0})
        ]

        self.TEST_UUID = '2bf5ddbb-c370-428f-9329-7379fc72488d'

    # def tearDown(self):
    #     pass

    def test_insert_single_record(self):
        ts = TimeSeries(name='TEST1', identifier=self.TEST_UUID, units='C', records=self.NUMBER_RECORDS)
        ts.insert(TsRecord(datetime.now(), 55))
        self.assertEqual(4, len(ts))
        # print(ts)
        #

    def test_insert_list_record(self):
        ts = TimeSeries(name='TEST1', identifier=self.TEST_UUID, units='C', records=self.NUMBER_RECORDS)
        ts.insert([TsRecord(datetime.now(), 55), TsRecord(datetime.now(), 77)])
        self.assertEqual(5, len(ts))

    # TODO N: test_to_JSON
    # def test_to_JSON(self):
    #     ts = TimeSeries(name='TEST1', identifier=self.TEST_UUID, units='C', records=self.NUMBER_RECORDS)
    #     ts_json = """{"id": "2bf5ddbb-c370-428f-9329-7379fc72488d", "name": "TEST1", "units": "C", "type": "NUMBER", "records": [{"timestamp": "%s", "value": 1, "quality": 192, "annotation": "comment"}, {"timestamp": "%s", "value": 2}]}""" % (
    #         self.ONE_MINUTE_AGO.isoformat(), self.NOW.isoformat())
    #     self.assertEqual(ts.toJSON(), ts_json)


class TestJtsDocument(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        self.NOW = datetime.now()  # new Date(Math.round(new Date().getTime() / 1000) * 1000)
        self.ONE_MINUTE_AGO = self.NOW - timedelta(minutes=1)  # new Date(NOW.getTime() - (60 * 1000))
        self.ONE_MINUTE_LATER = self.NOW + timedelta(minutes=1)  # new Date(NOW.getTime() - (60 * 1000))
        self.ONE_MILISECOND_AGO = self.NOW - timedelta(milliseconds=1)
        self.TWO_MILISECONDS_AGO = self.NOW - timedelta(milliseconds=2)
        self.THREE_MILISECONDS_AGO = self.NOW - timedelta(milliseconds=3)

        self.NUMBER_RECORDS = [
            TsRecord(**{"timestamp": self.ONE_MINUTE_AGO, "value": 1, "quality": 192, "annotation": 'comment'}),
            TsRecord(**{"timestamp": self.NOW, "value": 2}),
            TsRecord(**{"timestamp": self.TWO_MILISECONDS_AGO, "value": 0})]

        self.TEST_UUID = '2bf5ddbb-c370-428f-9329-7379fc72488d'

        self.NUMBER_SUBSECOND_RECORDS = [
            TsRecord(**{"timestamp": self.ONE_MINUTE_AGO, "value": 1, "quality": 192, "annotation": 'comment'}),
            TsRecord(**{"timestamp": self.THREE_MILISECONDS_AGO, "value": 2}),
            TsRecord(**{"timestamp": self.TWO_MILISECONDS_AGO, "value": 2}),
            TsRecord(**{"timestamp": self.ONE_MILISECOND_AGO, "value": 2}),
            TsRecord(**{"timestamp": self.NOW, "value": 2})]

        self.NUMBER_SUBSECOND_RECORDS_OUT_OF_ORDER = [
            TsRecord(**{"timestamp": self.ONE_MILISECOND_AGO, "value": 2}),
            TsRecord(**{"timestamp": self.NOW, "value": 2}),
            TsRecord(**{"timestamp": self.TWO_MILISECONDS_AGO, "value": 2}),
            TsRecord(**{"timestamp": self.ONE_MINUTE_AGO, "value": 1, "quality": 192, "annotation": 'comment'}),
            TsRecord(**{"timestamp": self.THREE_MILISECONDS_AGO, "value": 2})]

        '''
        Other sample dataTypes for tests:
        
        const TEXT_RECORDS = [
          { timestamp: ONE_MINUTE_AGO, value: 'Test value', quality: 192, annotation: 'comment' },
          { timestamp: NOW, value: null }
        ]

        const TIME_RECORDS = [
          { timestamp: ONE_MINUTE_AGO, value: ONE_MINUTE_AGO, quality: 192, annotation: 'comment' },
          { timestamp: NOW, value: null }
        ]

        const COORDINATES_RECORDS = [
          { timestamp: ONE_MINUTE_AGO, value: [100, 200], quality: 192, annotation: 'comment' },
          { timestamp: NOW, value: null }
        ]
        '''

    # def tearDown(self):
    #     pass

    def test_create_doc(self):
        jts_doc = JtsDocument(TimeSeries(name="series_1", records=self.NUMBER_RECORDS))
        self.assertEqual(len(jts_doc), 1)

    def test_add_single_series(self):
        jts_doc = JtsDocument()
        jts_doc.addSeries(TimeSeries(name="series_1", records=self.NUMBER_RECORDS))
        self.assertEqual(len(jts_doc), 1)

    def test_add_multiple_series(self):
        jts_doc = JtsDocument()
        jts_doc.addSeries([
            TimeSeries(name="series_3", records=self.NUMBER_RECORDS),
            TimeSeries(name="series_4", records=self.NUMBER_RECORDS)
        ])
        self.assertEqual(len(jts_doc), 2)

    def test_toJSON_NUMBER(self):
        timeseries1 = TimeSeries(identifier='series_1', name='Series 1', data_type='NUMBER', records=[
            TsRecord(**{"timestamp": self.ONE_MINUTE_AGO, "value": 1.23, "quality": 192, "annotation": 'comment'}),
            TsRecord(**{"timestamp": self.NOW, "value": 2.34, "quality": 245, "annotation": 'comment number 2'})]
                                 )
        timeseries2 = TimeSeries(identifier='series_2', name='Series 2', data_type='NUMBER', units="C", records=[
            TsRecord(**{"timestamp": self.NOW, "value": 1.11, "quality": 111, "annotation": 'comment ts2 111'}),
            TsRecord(**{"timestamp": self.ONE_MINUTE_AGO, "value": 0, "quality": 222, "annotation": 'comment ts2'}),
            TsRecord(**{"timestamp": self.ONE_MINUTE_LATER})
        ])

        jts_document = JtsDocument(series=[timeseries1, timeseries2])

        t = Template("""
        {
            "docType": "jts",
            "version": "1.0",
            "header": {
                "startTime": "$ONE_MINUTE_AGO",
                "endTime": "$NOW",
                "recordCount": 2,
                "columns": {
                    "0": {
                        "id": "series_1",
                        "name": "Series 1",
                        "dataType": "NUMBER"
                    },
                    "1": {
                        "id": "series_2",
                        "name": "Series 2",
                        "dataType": "NUMBER",
                        "units": "C"
                    }
                }
            },
            "data": [
                {
                    "ts": "$ONE_MINUTE_AGO",
                    "f": {
                        "0": {
                            "v": 1.23,
                            "q": 192,
                            "a": "comment"
                        },
                        "1": {
                            "v": 0,
                            "q": 222,
                            "a": "comment ts2"
                        }
                        
                    }
                },
                {
                    "ts": "$NOW",
                    "f": {
                        "0": {
                            "v": 2.34,
                            "q": 245,
                            "a": "comment number 2"
                        },
                        "1": {
                            "v": 1.11,
                            "q": 111,
                            "a": "comment ts2 111"
                        }
                    }
                }
            ]
        }
        """)

        jts_str = t.substitute(ONE_MINUTE_AGO=self.ONE_MINUTE_AGO.isoformat(timespec='milliseconds'),
                               NOW=self.NOW.isoformat(timespec='milliseconds'))
        jts_payload = jts_document.toJSONString()
        self.assertEqual(jts_payload, json.dumps(json.loads(jts_str)))

    def test_fromJSON(self):
        timeseries1 = TimeSeries(identifier='series_1', name='Series 1', data_type='NUMBER', records=[
            TsRecord(**{"timestamp": self.ONE_MINUTE_AGO, "value": 1.23, "quality": 192, "annotation": 'comment'}),
            TsRecord(**{"timestamp": self.NOW, "value": 2.34, "quality": 245, "annotation": 'comment number 2'})]
                                 )
        timeseries2 = TimeSeries(identifier='series_2', name='Series 2', data_type='NUMBER', units="C", records=[
            TsRecord(**{"timestamp": self.NOW, "value": 1.11, "quality": 111, "annotation": 'comment ts2 111'}),
            TsRecord(
                **{"timestamp": self.ONE_MINUTE_AGO, "value": 2.22, "quality": 222, "annotation": 'comment ts2 222'})]
                                 )

        jts_document = JtsDocument(series=[timeseries1, timeseries2])

        t = Template("""
        {
            "docType": "jts",
            "version": "1.0",
            "header": {
                "startTime": "$ONE_MINUTE_AGO",
                "endTime": "$NOW",
                "recordCount": 2,
                "columns": {
                    "0": {
                        "id": "series_1",
                        "name": "Series 1",
                        "dataType": "NUMBER"
                    },
                    "1": {
                        "id": "series_2",
                        "name": "Series 2",
                        "dataType": "NUMBER",
                        "units": "C"
                    }
                }
            },
            "data": [
                {
                    "ts": "$ONE_MINUTE_AGO",
                    "f": {
                        "0": {
                            "v": 1.23,
                            "q": 192,
                            "a": "comment"
                        },
                        "1": {
                            "v": 2.22,
                            "q": 222,
                            "a": "comment ts2 222"
                        }

                    }
                },
                {
                    "ts": "$NOW",
                    "f": {
                        "0": {
                            "v": 2.34,
                            "q": 245,
                            "a": "comment number 2"
                        },
                        "1": {
                            "v": 1.11,
                            "q": 111,
                            "a": "comment ts2 111"
                        }
                    }
                }
            ]
        }
        """)

        jts_str = t.substitute(ONE_MINUTE_AGO=self.ONE_MINUTE_AGO.isoformat(),
                               NOW=self.NOW.isoformat())
        jts_loaded = JtsDocument.fromJSON(jts_str)

        self.assertEqual(jts_loaded.toJSON(), jts_document.toJSON())

    def test_getSeriesByID(self):
        timeseries1 = TimeSeries(identifier='series_1', name='Series 1', data_type='NUMBER', records=[
            TsRecord(**{"timestamp": self.ONE_MINUTE_AGO, "value": 1.23, "quality": 192, "annotation": 'comment'}),
            TsRecord(**{"timestamp": self.NOW, "value": 2.34, "quality": 245, "annotation": 'comment number 2'})]
                                 )
        timeseries2 = TimeSeries(identifier='series_2', name='Series 2', data_type='NUMBER', units="C",
                                 records=TsRecord(timestamp=self.NOW, value=1.11, quality=111,
                                                  annotation='comment ts2 111')
                                 )
        jts_doc = JtsDocument([timeseries1, timeseries2])

        test_series = jts_doc.getSeries(identifier='series_2')
        self.assertEqual(test_series.identifier, "series_2")
        self.assertEqual(len(test_series), 1)

        # test('get series by id', () = > {
        #     const
        # jtsDocument = new
        # JtsDocument()
        # jtsDocument.addSeries([new TimeSeries({id: 'series_1', records: NUMBER_RECORDS}), new
        #                        TimeSeries({id: 'series_2', records: NUMBER_RECORDS})])
        # expect(jtsDocument.getSeries('series_1')?.length).toEqual(NUMBER_RECORDS.length)
        # })

        # TODO extend test to other dataTypes as below
        # describe('static methods', () => {
        #   test('construct from JSON', () => {
        #     expect(() => JtsDocument.from('')).toThrow()
        #     const jtsDocument = new JtsDocument()
        #     jtsDocument.addSeries(new TimeSeries({ type: 'NUMBER', records: NUMBER_RECORDS }))
        #     jtsDocument.addSeries(new TimeSeries({ type: 'TEXT', records: TEXT_RECORDS }))
        #     jtsDocument.addSeries(new TimeSeries({ type: 'TIME', records: TIME_RECORDS }))
        #     jtsDocument.addSeries(new TimeSeries({ type: 'COORDINATES', records: COORDINATES_RECORDS }))
        #     const jtsDocument2 = JtsDocument.from(jtsDocument.toJSON()) || new JtsDocument()
        #     expect(jtsDocument).toEqual(jtsDocument2)
        #     expect(jtsDocument.toString()).toEqual(jtsDocument2.toString())

    def test_subSecondRecords(self):
        jts_doc = JtsDocument()
        jts_doc.addSeries(TimeSeries(name="ts", identifier="series_1", records=self.NUMBER_SUBSECOND_RECORDS))
        self.assertEqual(len(jts_doc), 1)
        ts = jts_doc.getSeries(identifier="series_1")
        self.assertEqual(len(ts.records), 5)
        self.assertEqual(len(jts_doc.toJSON()['data']), 5)

    def test_sortingSubSecondRecords(self):
        expected = JtsDocument()
        expected.addSeries(TimeSeries(name="ts", identifier="series_1", records=self.NUMBER_SUBSECOND_RECORDS))

        jts_doc = JtsDocument()
        jts_doc.addSeries(
            TimeSeries(name="ts", identifier="series_1", records=self.NUMBER_SUBSECOND_RECORDS_OUT_OF_ORDER))

        self.assertEqual(len(jts_doc), 1)
        self.assertEqual(jts_doc.toJSON()['header']['startTime'], expected.toJSON()['header']['startTime'])
        self.assertEqual(jts_doc.toJSON()['header']['endTime'], expected.toJSON()['header']['endTime'])
        for i in range(5):
            self.assertEqual(jts_doc.toJSON()['data'][i]['ts'], expected.toJSON()['data'][i]['ts'])

        self.assertEqual(jts_doc.toJSON(), expected.toJSON())
        self.assertEqual(jts_doc.toJSONString(), expected.toJSONString())

        if __name__ == '__main__':
            unittest.main()

    def test_TimeSeries_with_str_value(self):
        # test if TimeSeries raises TypeError when value is not float or int

        with self.assertRaises(TypeError):
            timeseries1 = TimeSeries(identifier='series_1', name='Series 1', data_type='NUMBER', records=[
                TsRecord(**{"timestamp": self.NOW, "value": '2.34'})]
                                     )



import json
import uuid
from datetime import datetime
from typing import Union, List

from dateutil import parser

TimeSeriesDataType = ('NUMBER', 'TEXT', 'TIME', 'COORDINATES')


class CustomDatetimeConverter(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat(timespec='milliseconds')

        # if isinstance(obj, complex):
        #     return [obj.real, obj.imag]

        # Let the base class default method raise the TypeError
        return super().default(obj)


class TsRecord:
    """
    A record of TimeSeries object

    :param timestamp: Timestamp
    :type timestamp: datetime
    :param value: Value. For data type 'NUMBER', value must be int or float.
    :type value: Union[float, str, int]
    :param quality: Quality
    :type quality: int
    :param annotation: Annotation
    :type annotation: str
    """

    def __init__(self, timestamp: datetime, value: Union[float, str, int] = None, quality: int = None,
                 annotation: str = None):
        # TODO 1: enforce value types
        self.timestamp = timestamp
        self.value = value
        self.quality = quality
        self.annotation = annotation


class TimeSeries:
    """
    TimeSeries object

    :param data_type: Type of time series. E.g.: 'NUMBER', 'TEXT', 'TIME', 'COORDINATES'
    :type data_type: str, optional
    :default data_type: 'NUMBER'
    :param records: List of records
    :type records: list, optional
    :param name: Time series name
    :type name: str
    :param identifier: Time series ID. Autogenerated as UUID4 if not specified
    :type identifier: str, optional
    """

    def __init__(self, name: str, units: str = None, identifier: str = str(uuid.uuid4()),
                 data_type: str = 'NUMBER',
                 records: Union[List[TsRecord], TsRecord] = None):
        self.identifier = identifier
        self.name = name
        self.units = units
        self.data_type = data_type

        if records is None:
            self.records = []
        elif isinstance(records, TsRecord):
            self.records = [records]
        elif isinstance(records, list) and all(isinstance(x, TsRecord) for x in records):
            self.records = records
        else:
            raise TypeError("'records' value must be TsRecord or List[TsRecord]")

        for r in self.records:

            if (r.value is not None) and not (isinstance(r.value, (float, int))) and (self.data_type == 'NUMBER'):
                raise TypeError(
                    "TimeSeries with data type 'NUMBER' includes TSRecord with value '%s' which is %s. 'NUMBER' "
                    "values must be int or float" % (r.value, type(r.value))
                )

    # def __eq__(self, other):
    #     return self.__dict__ is other.__dict__

    def insert(self, records: Union[TsRecord, List[TsRecord]]):
        """
        Insert single or multiple records
        """

        if isinstance(records, list):
            self.records.extend(records)
        # single instance
        else:
            self.records.append(records)

    # def sort(self):
    #     pass
    #
    # def clone(self):
    #     # ITimeSeries<Type>;
    #     pass

    def __len__(self):
        return self.records.__len__()

    def toJSON(self) -> str:
        """
        Outputs formatted JSON
        """
        return json.dumps(self)


# TODO METHODS to implement
#
#   def get_timestamps:
#
#   public sort (): TimeSeries<Type> {
#
#   public clone (): TimeSeries<Type> {
#
#   private recordToJSON (record: ITimeSeriesRecord<Type>): ITimeSeriesRecordJson<Type> {
#
#   private valueToJSON (value: Type | undefined): Type | undefined | string | null {
#
#   private cloneRecords (records: ITimeSeriesRecord<Type>[]): ITimeSeriesRecord<Type>[] {


class JtsDocument:
    pass


class JtsDocument:
    """
    JTS document object

    :raise [TypeError]: [Value of 'series' must be types of TimeSeries or List[TimeSeries]]
    """

    """
    TODO Methods to implement

    // Clone document (also clones series)
    jtsDocument.clone()

    // Create a new jtsDocument from JSON
    const jtsDocument = JtsDocument.from('{"docType": "jts", ...}')

    Properties

    // Get JTS specification version number
    jtsDocument.version // 1

    // Get 
    jtsDocument.series // [timeseries1, timeseries2]
    """

    def __init__(self, series: Union[List[TimeSeries], TimeSeries] = None, version: str = "1.0"):
        # enforce accepted types
        if series is not None:
            if (not isinstance(series, list) and not isinstance(series, TimeSeries)) \
                    or (isinstance(series, list) and not all(isinstance(x, TimeSeries) for x in series)):
                raise TypeError("Value of 'series' must be types of TimeSeries or List[TimeSeries]")

        self.version = version

        if isinstance(series, list):
            self.series = series
        elif isinstance(series, TimeSeries):
            self.series = [series]
        else:
            self.series = []

    # def __eq__(self, other):
    #     return self.__dict__ is other.__dict__

    def addSeries(self, series: Union[List[TimeSeries], TimeSeries]):
        """
        Add single or multiple TimeSeries
        """
        if isinstance(series, list):
            self.series.extend(series)
        # single instance
        else:
            self.series.append(series)

    def __len__(self):
        return self.series.__len__()

    def toJSON(self) -> dict:
        """
        Output as dictionary of JSON structure

        :return: Python dictionary of JSON structure
        :rtype: dict
        """
        # return json.dumps(self.__build())
        return self.__build()

    def toJSONString(self) -> str:
        """
        Output as stringified JSON (json.dumps)

        :return: Output as stringified JSON
        :rtype: str
        """
        return json.dumps(self.toJSON(), cls=CustomDatetimeConverter)

    def __build(self):
        doc = dict(docType='jts',
                   version=self.version)

        data = self.__get_data()
        if not data:
            raise Exception("Cannot build without jts 'data'")
        header = self.__get_header(data)
        if header:
            doc['header'] = header
        doc['data'] = data

        return doc

    def __get_header(self, data):
        return dict(startTime=data[0]['ts'],
                    endTime=data[-1]['ts'],
                    recordCount=len(data),
                    columns=self.__getHeaderColumns()
                    ) if data else None

    def __getHeaderColumns(self):
        column_map = {}
        for idx, s in enumerate(self.series):
            column_map[idx] = dict(
                id=s.identifier,
                name=s.name,
                dataType=s.data_type,
            )
            if s.units:
                column_map[idx]["units"] = s.units

        return column_map

    # build "data" section of the document
    def __get_data(self):
        record_map = {}
        for idx, s in enumerate(self.series):
            for r in s.records:
                if (r.value is None) and (r.annotation is None) and (r.quality is None):
                    continue
                key = r.timestamp.timestamp()

                if not record_map.get(key):
                    # Dirty way to convert timestamp to string here, but it is to avoid datetime serialization upstream
                    record_map[key] = {"ts": r.timestamp.isoformat(timespec='milliseconds'), "f": {}}

                record_map[key]["f"][idx] = self.__getDataColumnFromRecord(r, s.data_type)  # dict of entry values

        # record_map_sorted = dict(sorted(record_map.items()))
        record_map_sorted = [x[1] for x in sorted(record_map.items())]  # use tuple here?

        return record_map_sorted

    def __getDataColumnFromRecord(self, r, data_type):

        column = {}
        v = r.value
        if v is not None:
            if data_type == 'NUMBER':
                if isinstance(v, float) or isinstance(v, int):
                    column["v"] = v
                else:
                    raise TypeError("Value of Data Type 'NUMBER' must be float or int")
            elif data_type == 'TEXT':
                column["v"] = str(v)

        # TODO other types below
        # case 'TIME': return { $time: (v as Date).toISOString?.() || 'invalid date' }
        # case 'COORDINATES': return { $coords: ((v as Array<number>).length === 2 ? v : []) }

        if r.quality:
            column["q"] = r.quality

        if r.annotation:
            column["a"] = r.annotation

        return column

    @staticmethod
    def fromJSON(json_str: str) -> JtsDocument:
        """
        Create a new jtsDocument from JSON
        """

        json_obj = json.loads(json_str)

        jts_doc = JtsDocument(version=json_obj.get('version'))

        # build series from header columns
        for idx, c in json_obj['header']["columns"].items():
            jts_doc.addSeries(TimeSeries(
                identifier=c.get("id"),
                name=c.get("name"),
                data_type=c.get("dataType"),
                units=c.get("units"))
            )

        # add records to corresponding series
        for e in json_obj['data']:

            ts = parser.parse(e["ts"])
            f = e["f"]

            for i, r in f.items():
                jts_doc.series[int(i)].insert(
                    TsRecord(
                        timestamp=ts,
                        value=r.get('v'),
                        quality=r.get('q'),
                        annotation=r.get('a'))
                )
                # else:
                #     raise Exception("Data columns do not match Header columns")

        return jts_doc

    def getSeries(self, identifier: str) -> TimeSeries:
        """
        Get series by id
        """

        # TODO: return list of found series
        return next((x for x in self.series if x.identifier == identifier), None)
